package org.roboquant.strategies

import org.roboquant.common.Asset
import org.roboquant.orders.LimitOrder
import org.roboquant.orders.MarketOrder

/**
 * Signal provides a [Rating] for an [Asset] and is typically created by a strategy.
 *
 * Besides, the asset and rating, a signal can also optionally provide the following information:
 *
 * - Take profit price for that asset
 * - Stop loss price for that asset
 * - The probability, or in other words how sure the strategy is about the signal
 *
 * It depends on the policy how these signals are translated into actual orders, but possible scenarios are:
 *
 *  - A BUY rating results in going long for that asset. A SELL rating results in going short
 *  - The target price will become an additional take profit order
 *  - The probability determines the stake (volume) of the order. The more confident, the larger the stake
 *
 * @property asset The asset for which this rating applies
 * @property rating The rating for the asset
 * @property takeProfit An optional take profit price in the same currency as the asset
 * @property stopLoss An optional stop loss price in the same currency as the asset
 * @property probability Optional the probability (value between 0.0 and 1.0) that the rating is correct.
 * @property source Optional source of the signal, like the strategy name
 * @constructor Create a new Signal
 */
class Signal(
    val asset: Asset,
    val rating: Rating,
    val takeProfit: Double = Double.NaN,
    val stopLoss: Double = Double.NaN,
    val probability: Double = Double.NaN,
    val source: String = ""
) {

    /**
     * Create a new signal and based on the strategy that created it
     */
    constructor(asset: Asset, rating: Rating, strategy: Strategy) : this(asset, rating, source = "$strategy")

    /**
     * Does this signal conflict with an [other] signal. Two signals conflict if they contain the same asset but opposite
     * ratings. So one signal has a positive outlook and the other one is negative.
     */
    fun conflicts(other: Signal) = asset == other.asset && rating.conflicts(other.rating)

    fun toMarketOrder(qty: Double) = MarketOrder(asset, qty, tag = source)

    fun toLimitOrder(qty: Double): LimitOrder {
        require(!takeProfit.isNaN()) { "Cannot create limit order since no take profit has been provided" }
        return LimitOrder(asset, qty, takeProfit, tag = source)
    }

}

/**
 * Different types of Signal Resolutions available
 */
enum class SignalResolution {

    /**
     * Don't resolve any potential signal conflicts
     */
    NONE,

    /**
     * Select first signal found for an asset
     */
    FIRST,

    /**
     * Select last signal found for an asset
     */
    LAST,

    /**
     * Select the first signal for an asset assuming there are no conflicting signals, see also [Signal.conflicts]
     */
    NO_CONFLICTS,

    /**
     * Select a signal for an asset is there are no duplicate signals, even if the duplicate signals are not
     * conflicting
     */
    NO_DUPLICATES,
}

/**
 * Resolve conflicting signals. For many strategies this might not be necessary since there is only 1 signal per
 * asset, but as strategies are combined, this issue might pop up. You can specify the resolution [rule] to apply
 * when solving conflicts, the default being [SignalResolution.NO_CONFLICTS].
 *
 * It returns the list of signals without any conflicts.
 */
fun List<Signal>.resolve(rule: SignalResolution = SignalResolution.NO_CONFLICTS): List<Signal> {
    if (size < 2) return this

    return when (rule) {
        SignalResolution.NONE -> this
        SignalResolution.FIRST -> distinctBy { it.asset }
        SignalResolution.LAST -> asReversed().distinctBy { it.asset }.asReversed()
        SignalResolution.NO_CONFLICTS -> filter { none { f -> f.conflicts(it) } }.distinctBy { it.asset }
        SignalResolution.NO_DUPLICATES -> {
            val assets = this.groupBy { it.asset }
            filter { assets[it.asset]!!.size == 1 }
        }
    }
}
